from __future__ import annotations

from pydantic_ai.exceptions import ApprovalRequired


class PolicyDenied(RuntimeError):
    def __init__(self, reason: str):
        self.reason = reason
        super().__init__(reason)


__all__ = ["ApprovalRequired", "PolicyDenied"]
